/* Nessus
 * Copyright (C) 2002 Michel Arboi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * $Id: nessus-identd.c,v 1.1 2002/07/25 20:24:25 mikhail Exp $
 */

#include <includes.h>

static int
takelock()
{
  char	lockname[MAXPATHLEN];
  int	fd;

  snprintf(lockname, MAXPATHLEN, "%s/identdlock", NESSUSD_STATEDIR);
  fd = open(lockname, O_WRONLY|O_CREAT|O_TRUNC, 0600);
  if (fd < 0)
    {
      perror(lockname);
      return -1;
    }
  if (lockf(fd, F_TLOCK, 1) < 0)
    {
      perror(lockname);
      (void) close(fd);
      return -1;
    }
  return fd;
}

static int
init_socket_tcp(int port)
{
  struct sockaddr_in	addr;


  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      perror("socket");
      return -1;
    }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sock, &addr, sizeof(addr)) < 0)
    {
      perror("bind");
      /* No use to close socket */
      return -1;
    }

  if (listen(sock, 10) < 0)
    {
      perror("listen");
      return -1;
    }

  return sock;
}

static int
init_socket_unix(const char* socketpath)
{
  struct sockaddr_un	addr;

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      perror("socket");
      return -1;
    }

  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;
  strcpy(addr.su_path, socketpath);
  if (bind(sock, &addr, sizeof(addr)) < 0)
    {
      perror("bind");
      /* No use to close socket */
      return -1;
    }

  if (listen(sock, 10) < 0)
    {
      perror("listen");
      return -1;
    }

  return sock;
}

#define MAX_CL	64	/* clients */

#define NI_RD	1
#define NI_RD_W	2
#define NI_WR	3
#define NI_ERR	4

struct {
  int		fd;		/* socket */
  short		st;		/* state */
  short		typ;		/* Type: 0 for Unix, 1 for remote */
  size_t	bufsz;		/* buffer size (malloced) */
  char		*buf, *pubf;	/* buffer & current R/W pointer */
}	cli[MAX_CL];

struct {
  int		pid;		/* pid */
  char		*name;		/* login name */
}	ids[MAX_CL];

static void
main_loop(int ts, int us)
{
  fd_set	rd, wr;
  int		nb = 0;
  int		i, j, n, maxfd, s, st, first_free = 0;
  char		line[80];


  /* Init */
  for (i = 0; i < MAX_CL; i ++)
    cli[i].fd = -1;

  /* Main loop */

  for (;;)
    {
      FD_ZERO(&rd); FD_ZERO(&wr);
      FD_SET(ts, &rd); FD_SET(us, &rd);
      maxfd = MAX(ts, us);

      for (i = 0; i < MAX_CL; i ++)
	if (cli[i].fd >= 0)
	  {
	    s = cli[i].fd; st = cli[i].st;
	    if (st == NI_RD)
	      FD_SET(s, &rd);
	    else if (st == NI_WR)
	      FD_SET(s, &wr); 
	    else 
	      s = -1;
	    if (s > maxfd)
	      maxfd = s;
	  }

      if (select(maxfd+1, &rd, &wr, NULL, NULL) < 0)
	{
	  perror("select");
	  return;
	}

      /* Incoming connections */
      
      if (FD_ISSET(us, &rd))
	{
	  if ((s = accept(us, NULL, NULL)) < 0)
	    perror("accept");
	  else
	    {
	      for (j = first_free; j < MAX_CL; j ++)
		if (cli[j].fd < 0)
		  break;
	      if (j < MAX_CL)
		{
		  cli[j].fd = s;
		  cli[j].st = NI_RD;
		  cli[j].pbuf = cli[j].buf;
		  first_free = j + 1;
		}
	      else
		close(s);	/* Dirty but simple ;-) */
	    }
	}
      /* Duplicated code */
      if (FD_ISSET(ts, &rd))
	{
	  if ((s = accept(ts, NULL, NULL)) < 0)
	    perror("accept");
	  else
	    {
	      for (j = first_free; j < MAX_CL; j ++)
		if (cli[j].fd < 0)
		  break;
	      if (j < MAX_CL)
		{
		  cli[j].fd = s;
		  cli[j].st = NI_RD;
		  cli[j].pbuf = cli[j].buf;
		  first_free = j + 1;
		}
	      else
		close(s);	/* Dirty but simple ;-) */
	    }
	}
      
      /* Established connections */

      for (i = 0; i < MAX_CL; i ++)
	{
	  s = cli[i].fd; st = cli[i].st;
	  if (s < 0)
	    continue;
	  if (st == NI_RD && FD_ISSET(s, &rd))
	    {
	      n = recv(s, line, sizeof(line), 0);
	      if (n < 0)
		{
		  perror("recv");
		  close(s);
		  cli[i].fd = -1;
		}
	    }
	}
    }
  
}

int
main(int argc, char* argv[])
{
  int		fd, port, s1, s2;
  char		*socketpath;
  
  if (argc != 3)
    {
      fprintf(stderr, "This program should be run by nessusd\n\
Usage: nessus-identd port socketpath\n");
      return 1;
    }

  if ((fd = takelock()) < 0)
    {
      fprintf(stderr, "Could not take lock\n");
      return 1;
    }
  
  port = atoi(argv[1]);
  socketpath = argv[2];

  if ((s1 = init_socket_tcp(port)) < 0)
    return 1;			/* Will close the file & release the lock */
  if ((s2 = init_socket_unix(socketpath)) < 0)
    return 1;			/* Will close the file & socket */
  main_loop(s1, s2);
  return 0;
}
